// (C) 2005  Willem Hengeveld <itsme@xs4all.nl>
//
#include "pro.h"     // for basic types.
#include "ida.hpp"   // for ida constants and types.
#include "netnode.hpp"  // for RootNode

#include "expr.hpp"  // for IDCFuncs

#include <fstream>
#include <iostream>
#include <iomanip>
#include <cassert>
#include <string>

#include "boost/format.hpp"

#include <windows.h>
#include <tlhelp32.h>
typedef struct {
    char  *function_name;      // 00
    char  *end_function_name;  // 04
    ulong w1;                  // 08
    ulong w2;                  // 0c
    uchar *after_end_fnname;   // 10
    ulong w4;                  // 14
    ulong nrparams;            // 18
    uchar *body;               // 1c
    uchar *end_body;           // 20
    ulong w5;                  // 24
    ulong w6;                  // 28
    ulong w7;                  // 2c
    ulong w8;                  // 30
} compiled_func_t;


// !!! GLOBAL ... initialized by dump_compiled_functions
//   used by disassemble, to resolve compiled functions indexes.
compiled_func_t *g_flist;




std::string hexdump(const uchar *p, int len)
{
    if (p==NULL)
        return "(null)";
    std::string str;
    for (int i=0 ; i<len ; i++) {
        if (!str.empty()) str += " ";
        char strbuf[16];
        qsnprintf(strbuf, 16, "%02x", p[i]);

        str += strbuf;
    }
    return str;
}
std::string ascdump(const char *p, size_t len)
{
    for (size_t i=0 ; p[i] && i<len ; i++) {
        // -- find end of string
    }
    return std::string(p, i);
}
std::string tagstr(char tag)
{
    if (tag>=' ' && tag<='~') {
        char strbuf[2];
        strbuf[0]= tag;
        strbuf[1]= 0;
        return std::string(strbuf);
    }
    else {
        char strbuf[16];
        qsnprintf(strbuf, 16, "%02x", (uchar)tag);
        return std::string(strbuf);
    }
}
std::ostream& operator<<(std::ostream& os, nodeidx_t idx)
{
    return os << boost::format("%08lx") % ((int)idx);
}
std::ostream& operator<<(std::ostream& os, netnode& node)
{
    os << boost::format("node %08lx ll=%d  %s\n") 
            % ((nodeidx_t)node) 
            % node.last_length()
            % (node.name()?node.name():"(null)");

    if (node.value_exists()) {
        os << boost::format("long_value: %08lx str=%s\n") 
            % node.long_value()
            % ascdump(node.value(), MAXSPECSIZE);
    }
    else {
        os << "no value\n";
    }
    for (int tag= 0 ; tag < 0x100 ; tag++)
        for (nodeidx_t idx= node.sup1st(tag) ; idx!=BADNODE ; idx= node.supnxt(idx, tag))
        {
            os << boost::format("[%s, %08lx] : alt=%08lx ch=%02x sup=%s\n") 
                    % tagstr(tag) % idx
                    % node.altval(idx, tag)
                    % ((int)node.charval(idx, tag))
                    % ascdump(node.supval(idx, tag), MAXSPECSIZE);
        }

    for (int tag=0 ; tag<0x100 ; tag++)
        for (char* idx= node.hash1st() ; idx ; idx=node.hashnxt(idx, tag))
            os << boost::format("hashval(%s, %s) : %s\n")
                    % tagstr(tag) % idx
                    % (node.hashval(idx, tag)?node.hashval(idx, tag):"(null)");
/* ... netlink is missing from ida.lib
    netlink l;
    if (l.start())
        os << boost::format("listing links of type %s\n") % l.name();
        do {
            for (netnode n= l.firstlink(node); n!=BADNODE ; n= l.nextlink(node, n)) {
                os << boost::format("  link to %08lx %s\n") % (nodeidx_t)n % n.name();
            }
        } while (l.next());
*/
    return os;
}
std::string functionargs_string(const char *args)
{
    std::string str;
    while (*args) {
        if (!str.empty()) str += ", ";
        switch(*args)
        {
            case VT_STR:   str += "string"; break;
            case VT_LONG:  str += "long"; break;
            case VT_FLOAT: str += "float"; break;
            case VT_WILD:  str += "..."; break;
            default: 
               char strbuf[16];
               qsnprintf(strbuf, 16, "VT_%08lx", *args);
               str += strbuf;
        }

        args++;
    }
    return str;
}
void disassemble(std::ostream&os, const uchar *body, int len)
{
    for (int i=0 ; i<len ; i++)
    {
        os << boost::format("%08lx %04x  %02x")
            % ((int)body+i)
            % i
            % ((int)body[i]);
        int nyble = body[i]&0xf;
        int byte  = (i+1<len)?body[i+1] : 0;
        int word  = (i+2<len)?(body[i+2]<<8)|body[i+1]:0;
        int dword = (i+4<len)?(body[i+4]<<24)|(body[i+3]<<16)|(body[i+2]<<8)|body[i+1]:0;

        switch(body[i]&0xf0) {
            case 0x00: // call internfunc 00-0f
                os << boost::format("          ; call %04x %s\n")
                    % nyble
                    % IDCFuncs.f[nyble].name;
                break;
            case 0x10: // call userdef func 00-0f
                os << boost::format("          ; call compiled func %02x %s\n")
                    % nyble
                    % g_flist[nyble].function_name;
                break;
            case 0x20: // pop var 00-0f
                os << boost::format("          ; pop var%02x\n")
                    % nyble;
                break;
            case 0x30: // pop param 00-0f
                os << boost::format("          ; pop param%02x\n")
                    % nyble;
                break;
            case 0x40: // var 00-0f = tos
                os << boost::format("          ; var%02x := tos\n")
                    % nyble;
                break;
            case 0x50: // param 00-0f = tos
                os << boost::format("          ; param%02x := tos\n")
                    % nyble;
                break;
            case 0x60: // push var 00-0f
                os << boost::format("          ; push var%02x\n")
                    % nyble;
                break;
            case 0x70: // push param 00-0f
                os << boost::format("          ; push param%02x\n")
                    % nyble;
                break;
            // case 0x80: these are handled in default:
            // case 0x90: unknown
            // case 0xa0: these are handled in default:
            // case 0xb0: these are handled in default:
            // case 0xc0: these are handled in default:
            // case 0xd0: these are handled in default:
            case 0xe0:
                os << boost::format("          ; push #%02x\n")
                    % nyble;
                break;
            // case 0xf0: unknown

            // handle non nyble oriented opcodes
            default: switch(body[i]) {
            case 0x80:
                if (i >= len-1) { os << boost::format("\nERROR: unexpected end of code\n"); break; }
                os << boost::format(" %02x       ; call %04x %s\n")
                    % byte
                    % byte
                    % IDCFuncs.f[byte].name;
                i++;
                break;
            case 0x81:
                if (i >= len-2) { os << boost::format("\nERROR: unexpected end of code\n"); break; }
                os << boost::format(" %04x     ; call %04x %s\n")
                    % word
                    % word
                    % IDCFuncs.f[word].name;
                i += 2;
                break;
            case 0x82:
                if (i >= len-1) { os << boost::format("\nERROR: unexpected end of code\n"); break; }
                os << boost::format(" %02x       ; call compiled func %02x %s\n")
                    % byte
                    % byte
                    % g_flist[byte].function_name;
                i++;
                break;
            case 0x83:
                if (i >= len-2) { os << boost::format("\nERROR: unexpected end of code\n"); break; }
                os << boost::format(" %04x     ; call compiled func %04x %s\n")
                    % word
                    % word
                    % g_flist[word].function_name;
                i += 2;
                break;
            case 0x84:
                if (i >= len-1) { os << boost::format("\nERROR: unexpected end of code\n"); break; }
                os << boost::format(" %02x       ; pop var%02x\n")
                    % byte
                    % byte;
                i++;
                break;
            // case 0x85:  unknown
            case 0x86:
                if (i >= len-1) { os << boost::format("\nERROR: unexpected end of code\n"); break; }
                os << boost::format(" %02x       ; param%02x := tos\n")
                    % byte
                    % byte;
                i++;
                break;
            // case 0x87:  unknown
            case 0x88:
                if (i >= len-1) { os << boost::format("\nERROR: unexpected end of code\n"); break; }
                os << boost::format(" %02x       ; var%02x := tos\n")
                    % byte
                    % byte;
                i++;
                break;
            // case 0x89:  unknown
            case 0x8a:
                if (i >= len-1) { os << boost::format("\nERROR: unexpected end of code\n"); break; }
                os << boost::format(" %02x       ; pop param%02x\n")
                    % byte
                    % byte;
                i++;
                break;
            // case 0x8b:  unknown
            case 0x8c:
                if (i >= len-1) { os << boost::format("\nERROR: unexpected end of code\n"); break; }
                os << boost::format(" %02x       ; push var%02x\n")
                    % byte
                    % byte;
                i++;
                break;
            // case 0x8d:  unknown
            case 0x8e:
                if (i >= len-1) { os << boost::format("\nERROR: unexpected end of code\n"); break; }
                os << boost::format(" %02x       ; push param%02x\n")
                    % byte
                    % byte;
                i++;
                break;
            // case 0x8f:  unknown
            case 0xa0:
                if (i >= len-1) { os << boost::format("\nERROR: unexpected end of code\n"); break; }
                os << boost::format(" %02x       ; push #%02x\n")
                    % byte
                    % byte;
                i++;
                break;
            case 0xa1:
                if (i >= len-2) { os << boost::format("\nERROR: unexpected end of code\n"); break; }
                os << boost::format(" %04x     ; push #%04x\n")
                    % word
                    % word;
                i+=2;
                break;
            // case 0xa2:  unknown
            // case 0xa3:  unknown
            // case 0xa4:  unknown
            case 0xa5:
                if (i >= len-2) { os << boost::format("\nERROR: unexpected end of code\n"); break; }
                os << boost::format(" %04x     ; allocate %d local variables\n")
                    % word
                    % word;
                i+=2;
                break;
            // case 0xa6:  unknown
            case 0xa7:
                if (i >= len-4) { os << boost::format("\nERROR: unexpected end of code\n"); break; }
                os << boost::format(" %08x ; push #%08x\n")
                    % dword
                    % dword;
                i+=4;
                break;
            case 0xa8:
                if (i >= len-2-word) { os << boost::format("\nERROR: unexpected end of code\n"); break; }
                os << boost::format(" %04x ... ; push \"%s\"\n")
                    % word
                    % std::string((char*)body+i+3, word).c_str();
                i+=word+2;
                break;
            case 0xa9:
                os << boost::format("          ; pop\n");
                break;
            case 0xaa:
                os << boost::format("          ; start function param list\n");
                break;
            case 0xab:
                if (i >= len-4) { os << boost::format("\nERROR: unexpected end of code\n"); break; }
                os << boost::format(" %08x ; branch always %08x -> %04x\n")
                    % dword
                    % dword
                    % (i+dword+5);
                i+=4;
                break;
            case 0xac:
                if (i >= len-4) { os << boost::format("\nERROR: unexpected end of code\n"); break; }
                os << boost::format(" %08x ; branch do_while true %08x -> %04x\n")
                    % dword
                    % dword
                    % (i+dword+5);
                i+=4;
                break;
            case 0xad:
                if (i >= len-4) { os << boost::format("\nERROR: unexpected end of code\n"); break; }
                os << boost::format(" %08x ; branch if true %08x -> %04x\n")
                    % dword
                    % dword
                    % (i+dword+5);
                i+=4;
                break;
            case 0xae:
                if (i >= len-2-word) { os << boost::format("\nERROR: unexpected end of code\n"); break; }
                os << boost::format(" %04x ... ; call by name \"%s\"\n")
                    % word
                    % std::string((char*)body+i+3, word).c_str();
                i+=word+2;
                break;
            case 0xaf: os << boost::format("          ; return\n"); break;
            case 0xb0: os << boost::format("          ; add\n"); break;
            case 0xb1: os << boost::format("          ; substract\n"); break;
            case 0xb2: os << boost::format("          ; multiply\n"); break;
            case 0xb3: os << boost::format("          ; divide\n"); break;
            case 0xb4: os << boost::format("          ; modulus\n"); break;
            case 0xb5: os << boost::format("          ; logical or\n"); break;
            case 0xb6: os << boost::format("          ; logical and\n"); break;
            case 0xb7: os << boost::format("          ; logical not\n"); break;
            case 0xb8: os << boost::format("          ; bitwise or\n"); break;
            case 0xb9: os << boost::format("          ; bitwise and\n"); break;
            case 0xba: os << boost::format("          ; bitwise xor\n"); break;
            case 0xbb: os << boost::format("          ; unary bitwise negate\n"); break;
            case 0xbc: os << boost::format("          ; shiftright\n"); break;
            case 0xbd: os << boost::format("          ; shiftleft\n"); break;
            case 0xbe: os << boost::format("          ; unary minus\n"); break;
            // case 0xbf:  unknown
            case 0xc0:
                if (i >= len-1) { os << boost::format("\nERROR: unexpected end of code\n"); break; }
                os << boost::format(" %02x       ; increment var%02x\n")
                    % byte
                    % byte;
                i++;
                break;
            // case 0xc1:  unknown
            case 0xc2:
                if (i >= len-1) { os << boost::format("\nERROR: unexpected end of code\n"); break; }
                os << boost::format(" %02x       ; increment param%02x\n")
                    % byte
                    % byte;
                i++;
                break;
            // case 0xc3:  unknown
            case 0xc4:
                if (i >= len-1) { os << boost::format("\nERROR: unexpected end of code\n"); break; }
                os << boost::format(" %02x       ; decrement var%02x\n")
                    % byte
                    % byte;
                i++;
                break;
            // case 0xc5:  unknown
            case 0xc6:
                if (i >= len-1) { os << boost::format("\nERROR: unexpected end of code\n"); break; }
                os << boost::format(" %02x       ; decrement param%02x\n")
                    % byte
                    % byte;
                i++;
                break;
            // case 0xc7:  unknown
            case 0xc8: os << boost::format("          ; compare ==\n"); break;
            case 0xc9: os << boost::format("          ; compare !=\n"); break;
            case 0xca: os << boost::format("          ; compare <=\n"); break;
            case 0xcb: os << boost::format("          ; compare >=\n"); break;
            case 0xcc: os << boost::format("          ; compare <\n"); break;
            case 0xcd: os << boost::format("          ; compare >\n"); break;
            case 0xce: os << boost::format("          ; cv to integer\n"); break;
            case 0xcf: os << boost::format("          ; cv to string\n"); break;
            case 0xd0: os << boost::format("          ; cv to float\n"); break;
            case 0xd1: 
                if (i >= len-12) { os << boost::format("\nERROR: unexpected end of code\n"); break; }
                os << boost::format(" .......... ; push float %s\n")
                    % hexdump(body+i+1, 12);
                i+=12;
                break;
            // case 0xd2 .. 0xdf:  unknown
            default:
                os << boost::format("            ; ERROR: unknown opcode\n");
             }
        }
    }
}
void funcbody(std::ostream&os, const char *name)
{
    int nargs=0;
    size_t bodylen=0;
    uchar *body= get_idc_func_body(name, &nargs, &bodylen);
    if (body)
    {
        os << boost::format("body: %08lx-%08lx: %s(%d)\n")
            % ((int)body)
            % ((int)body+bodylen)
            % name
            % nargs;
        disassemble(os, body, bodylen);
        os << "\n";
    }
}
std::ostream& operator<<(std::ostream& os, const extfun_t& f)
{
    funcbody(os, f.name);

    return os << boost::format("fu=%08lx fl=%08lx %s(%s)")
        % ((int)f.fp)
        % f.flags
        % f.name
        % functionargs_string(f.args);
}
std::ostream& operator<<(std::ostream& os, funcset_t& fset)
{
    os << boost::format("%d functions\n") % fset.qnty;
    for (int i=0 ; i<fset.qnty ; i++) {
        os << boost::format("function %d : %s\n") % i % fset.f[i];
    }
    return os;
}

void dump_idc_funcs(std::ostream& os)
{
    typedef struct {
        compiled_func_t *start;
        compiled_func_t *end;
    } startendpair_t;
    startendpair_t *flist= (startendpair_t *)0x100EC1B0;

    g_flist= flist->start;
    for (compiled_func_t *f= g_flist ; f<flist->end ; f++) {
        os << boost::format("body: %08lx-%08lx: %s(%d)\n")
            % ((int)f->body)
            % ((int)f->end_body)
            % f->function_name
            % ((int)f->nrparams);
        disassemble(os, f->body, f->end_body-f->body);
    }
}

void dump_db()
{
    std::filebuf fb; fb.open("dump_db.log", std::ios::out);
    std::ostream os(&fb);

    os << "------------rootnode-----------\n";
    os << RootNode;

    os << "------------node.start.next-----------\n";
    netnode n; 
    if (n.start()) 
        do {
            os << n; 
        } while (n.next());

    os << "------------idcfuncs-----------\n";
    os << IDCFuncs;

    // note: there seems to be no official way of getting a list of
    // currently loaded idc functions.

    os << "------------compiled functions-----------\n";

    // most of these functions are from
    //   http://www.xs4all.nl/~itsme/projects/disassemblers/ida.html
    //   http://nah6.com/~itsme/cvs-xdadevtools/idcscripts/formatdata.idc

    dump_idc_funcs(os);
}
